/*
 * file: Pv.c
 * This file is part of the libpv library.
 *
 * Copyright 2005 Pengutronix
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 */

#include <stdio.h>

#include <pv_register.h>

#include "Pv.h"
#include "PvEngine.h"

PyObject *Pv_new(PyTypeObject *type, PyObject *args, PyObject *keywds)
{
	PvObject *self;
	char *name;
	PvEngineObject *engine;
	static char *kwlist[] = { "engine", "name", NULL };

	//printf("DEBUG: %s\n", __FUNCTION__);

	if (!PyArg_ParseTupleAndKeywords(args, keywds, "O!s", kwlist,
			&PvEngine_Type, &engine, &name))
		return NULL;

	self = (PvObject*)type->tp_alloc(type, 0);
	if (!self)
		return NULL;

	self->engine = (struct pv_engine*)&engine->engine;
	if ((self->pv = pv_register_lookup(self->engine, name)) == 0) {
		PyErr_SetString(PyExc_AttributeError, "pv_register_lookup failed");
		Py_DECREF(self);
		return NULL;
	}

	return (PyObject *)self;
}


PyObject *Pv_get(PvObject *self, PyObject *args)
{
	//printf("DEBUG: %s\n", __FUNCTION__);

	switch (self->pv->type) {
	case PV_TYPE_INT8:
		return PyInt_FromLong(*pv_to_abs_addr(self->engine, self->pv)->s8);
	case PV_TYPE_UINT8:
		return PyInt_FromLong(*pv_to_abs_addr(self->engine, self->pv)->u8);
	case PV_TYPE_INT16:
		return PyInt_FromLong(*pv_to_abs_addr(self->engine, self->pv)->s16);
	case PV_TYPE_UINT16:
		return PyInt_FromLong(*pv_to_abs_addr(self->engine, self->pv)->u16);
	case PV_TYPE_INT32:
		return PyInt_FromLong(*pv_to_abs_addr(self->engine, self->pv)->s32);
	case PV_TYPE_UINT32:
		/* The python integer cannot hold "uint32_t" values. */
		return PyLong_FromUnsignedLong(*pv_to_abs_addr(self->engine, self->pv)->u32);
	case PV_TYPE_FLOAT:
		return PyFloat_FromDouble(*pv_to_abs_addr(self->engine, self->pv)->f);
	case PV_TYPE_DOUBLE:
		return PyFloat_FromDouble(*pv_to_abs_addr(self->engine, self->pv)->d);
	case PV_TYPE_STRING:
		return PyString_FromString(pv_to_abs_addr(self->engine, self->pv)->string);
	default:
		Py_INCREF(Py_None);
		return Py_None;
	};
}

PyObject *Pv_set(PvObject *self, PyObject *args)
{
	int64_t var;
	double fvar;
	char *svar;
	//printf("DEBUG: %s\n", __FUNCTION__);
	switch (self->pv->type) {
	case PV_TYPE_INT8:
		if (!PyArg_ParseTuple(args, "b", &var))
			return NULL;
		*pv_to_abs_addr(self->engine, self->pv)->s8 = var;
		break;
	case PV_TYPE_UINT8:
		if (!PyArg_ParseTuple(args, "B", &var))
			return NULL;
		*pv_to_abs_addr(self->engine, self->pv)->u8 = var;
		break;
	case PV_TYPE_INT16:
		if (!PyArg_ParseTuple(args, "h", &var))
			return NULL;
		*pv_to_abs_addr(self->engine, self->pv)->s16 = var;
		break;
	case PV_TYPE_UINT16:
		if (!PyArg_ParseTuple(args, "H", &var))
			return NULL;
		*pv_to_abs_addr(self->engine, self->pv)->u16 = var;
		break;
	case PV_TYPE_INT32:
		if (!PyArg_ParseTuple(args, "i", &var))
			return NULL;
		*pv_to_abs_addr(self->engine, self->pv)->s32 = var;
		break;
	case PV_TYPE_UINT32:
		if (!PyArg_ParseTuple(args, "I", &var))
			return NULL;
		*pv_to_abs_addr(self->engine, self->pv)->u32 = var;
		break;
	case PV_TYPE_FLOAT:
		if (!PyArg_ParseTuple(args, "f", &fvar))
			return NULL;
		*pv_to_abs_addr(self->engine, self->pv)->f = fvar;
		break;
	case PV_TYPE_DOUBLE:
		if (!PyArg_ParseTuple(args, "d", &fvar))
			return NULL;
		*pv_to_abs_addr(self->engine, self->pv)->d = fvar;
		break;
	case PV_TYPE_STRING:
		if (!PyArg_ParseTuple(args, "s", &svar))
			return NULL;
		pv_set_string((pv_engine_t*)self->engine,
			pv_to_abs_addr(self->engine, self->pv)->string, svar);
		break;
	default:
		break;
	};
	Py_INCREF(Py_None);
	return Py_None;
}

PyMethodDef Pv_methods[] = {
	{"set",	(PyCFunction)Pv_set,	METH_VARARGS,	PyDoc_STR("set...")},
	{"get",	(PyCFunction)Pv_get,	METH_VARARGS,	PyDoc_STR("get...")},
	{NULL},
};

void
Pv_dealloc(PvObject *self)
{
	//printf("DEBUG: %s\n", __FUNCTION__);
	PyObject_Del(self);
}

PyObject *Pv_str(PvObject *self)
{
	PyObject *o =  Pv_get(self, 0);
	PyObject *ret = o->ob_type->tp_str(o);
	Py_DECREF(o);
	return ret;
}

PyTypeObject Pv_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)	/*ob_type*/
	0,				/*ob_size*/
	"libpv.Pv",		/*tp_name*/
	sizeof(PvObject),		/*tp_basicsize*/
	0,				/*tp_itemsize*/
	/* methods */
	(destructor)Pv_dealloc,	/*tp_dealloc*/
	0,				/*tp_print*/
	0,				/*tp_getattr*/
	0,				/*tp_setattr*/
	0,				/*tp_compare*/
	0,				/*tp_repr*/
	0,				/*tp_as_number*/
	0,				/*tp_as_sequence*/
	0,				/*tp_as_mapping*/
	0,				/*tp_hash*/
	0,				/*tp_call*/
	(reprfunc)Pv_str,		/*tp_str*/
	0,				/*tp_getattro*/
	0,				/*tp_setattro*/
	0,				/*tp_as_buffer*/
	Py_TPFLAGS_DEFAULT,		/*tp_flags*/
	0,				/*tp_doc*/
	0,				/*tp_traverse*/
	0,				/*tp_clear*/
	0,				/*tp_richcompare*/
	0,				/*tp_weaklistoffset*/
	0,				/*tp_iter*/
	0,				/*tp_iternext*/
	Pv_methods,			/*tp_methods*/
	0,				/*tp_members*/
	0,				/*tp_getset*/
	0,				/*tp_base*/
	0,				/*tp_dict*/
	0,				/*tp_descr_get*/
	0,				/*tp_descr_set*/
	0,				/*tp_dictoffset*/
	0,				/*tp_init*/
	0,				/*tp_alloc*/
	(newfunc)Pv_new,		/*tp_new*/
	0,				/*tp_free*/
	0,				/*tp_is_gc*/
};

