/*
 * file: pv_init.c
 * This file is part of the libpv library.
 *
 * Copyright 2005, 2009 Pengutronix
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 */

#include <libpv_config.h>

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>

#include <libpv.h>
#include <pv_init.h>

#include <pv_bs.h>
#include <pv_bs_8bit.h>
#include <pv_lock.h>
#include <pv_destroy.h>
#include <pv_jvisu.h>
#include <pv_ev.h>		/* only used, when event handling is enabled */

#ifdef DEBUG
int _libpv_debuglevel = LEVEL_ALL;
#else
int _libpv_debuglevel = LEVEL_NOTICE;
#endif


static void pv_init_show_pvs(struct pv_engine	*engine)
{
	struct pv	*pv;
	int		finished = 0;

	printf("process variables:\n");
	printf("------------------\n\n");

	for (goto_head(&engine->pv_list); !finished; goto_next(&engine->pv_list)) {
		pv = (struct pv *)list_get_element_data(engine->pv_list);

		if (is_last(engine->pv_list))
			finished = 1;

		printf("id \"%s\", offset %d, type %d, size %zd, arraylen %zd\n",
		       pv->id, (int)pv->offset, pv->type, pv->size, pv->arraylen);
	}

	printf("\n");
}


static size_t pv_init_get_pvs_size(struct pv_list	*pv_list)
{
	size_t	size = 0;
	int	finished = 0;

	for (goto_head(&pv_list); !finished; goto_next(&pv_list)) {
		struct pv *pv = (struct pv *)list_get_element_data(pv_list);

		if (is_last(pv_list))
			finished = 1;

		size = MAX(size, pv->offset + (pv->size * pv->arraylen));
	}

	dprintf(LEVEL_INFO, "(%s) store size %zd\n",
		__FUNCTION__, size);

	return size;
}

/**
 * pv_init - Initialize the whole process image engine
 * @engine_anon: Current process image to work for
 * @attr: Specifies this process image
 *
 * Description
 *
 * Return value: EINVAL in case of parameter errors (NULL pointers, empty strings)
 **/
int pv_init(pv_engine_t		*engine_anon,
	    pv_engine_attr_t	*attr)
{
	struct pv_engine	*engine = (struct pv_engine *)engine_anon;
	int			err, ret;
	size_t			size;

	if (!engine || !attr)
		return -EINVAL;

	if (attr->desc_filename[0] == 0 ||
	    attr->store_filename[0] == 0)
		return -EINVAL;

	memset(engine, 0, sizeof(struct pv_engine));
	/*
	 * the shared resource API has strangly return values
	 * in case of error. shmat() for example -1 instead of NULL
	 * for an invalid pointer. To accommodate this
	 * we setup matching default values here
	 */
#ifdef CONFIG_EVENT
	engine->mqdes=(mqd_t)-1;	/* mq_open() */
	engine->mq_notify=(mqd_t)-1;	/* mq_open() */
	engine->event_server_key=-1;	/* ftok() */
#endif	/* CONFIG_EVENT */

	/* read pv from xml into list */
	err = pv_init_read_pvs(engine, attr->desc_filename);
	if (err)
		goto exit_0;

	/* show the pvs we've found */
	if (_libpv_debuglevel >= LEVEL_DEBUG)
		pv_init_show_pvs(engine);


	/* find out size of pv area */
	size = pv_init_get_pvs_size(engine->pv_list);
	assert(size != 0);


	/* init locking */
	err = pv_lock_init(engine, attr->store_filename);
	if (err)
		goto exit_1;

	/* init the backing store */
	err = pv_bs_init(engine, attr->store_type, attr->store_filename, size);
	if (err)
		goto exit_1;	/* FIXME: destroy locking (mkl) */

#ifdef CONFIG_EVENT
	if (!attr->is_server) {
	/* add event handling for this instance */
		err = pv_event_init_handling(engine,attr->store_filename);
	/*
	 * with event notification support it would be an error
	 * if there is no event daemon in the background. But if
	 * none want to use any notifications its all right!
	 * Changed from 1.2.0
	 *
	 *	if (err)
	 *		goto exit_2;
	 */
	}
#endif

	/* init jvisu */
	err = pv_jvisu_init(engine, attr->jvisu_port);
	if (err)
		goto exit_2;	/* FIXME: destroy locking (mkl) */

	return 0;


 exit_2:
	/* remove event handling for this instance */
	pv_event_exit_handling(engine);

	ret = pv_bs_destroy(engine);
	if (ret) {
		err = ret;
		dprintf(LEVEL_ERROR, "(%s) failed to destroy backing store, err %d\n",
			__FUNCTION__, err);
	}
 exit_1:
	ret = pv_destroy_free_pvs(engine->pv_list);
	if (ret) {
		err = ret;
		dprintf(LEVEL_ERROR, "(%s) failed to destroy pv_list, err %d\n",
			__FUNCTION__, err);
	}
 exit_0:
	return err;
}

/* end of file pv_init.c */
