/*
 * file: pv_init_expat.c
 * This file is part of the libpv library.
 *
 * Copyright 2009 Jens Stadelmann <jst@pengutronix.de>
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 */

#include <libpv_config.h>

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>

#include <expat.h>

#include <libpv.h>

#include <pv_init.h>
#include <pv_destroy.h>

#define XMLBUFFERSIZE	(255)

enum pv_xml_elem {
	PV_XML_ELEM_TYPE,
	PV_XML_ELEM_SIZE,
	PV_XML_ELEM_REPEAT,
	PV_XML_ELEM_PV,
	PV_XML_ELEM_OFFSET,
	PV_XML_ELEM_NAME,
	PV_XML_ELEM_MIN,
	PV_XML_ELEM_MAX,
	PV_XML_ELEM_DEVICE,	/* not supported! */
	PV_XML_ELEM_UNDEF,
};

/* The tag names must be in reversed lexicographical order for later comparisons. */
const char *pv_xml_tag[] = {
	[PV_XML_ELEM_TYPE] = "type",
	[PV_XML_ELEM_SIZE] = "size",
	[PV_XML_ELEM_REPEAT] = "repeat",
	[PV_XML_ELEM_PV] = "pv",
	[PV_XML_ELEM_OFFSET] = "offset",
	[PV_XML_ELEM_NAME] = "name",
	[PV_XML_ELEM_MIN] = "min",
	[PV_XML_ELEM_MAX] = "max",
	[PV_XML_ELEM_DEVICE] = "device",
};

struct pv_xml_parse_data {
	int err;
	int ignore;		/* flag used to ignore unknown elements inside <pv> ... </pv> */
	enum pv_xml_elem cur_elem;	/* content of this element is currently read */
	char *buf;
	struct pv *pv_ptr;
	struct pv_list *pv_list;
	XML_Parser *parser;
};

static enum pv_type pv_xml_get_type(const char *ts)
{
	if (!strcmp(ts, "uint8_t"))
		return PV_TYPE_UINT8;
	else if (!strcmp(ts, "uint16_t"))
		return PV_TYPE_UINT16;
	else if (!strcmp(ts, "uint32_t"))
		return PV_TYPE_UINT32;
	else if (!strcmp(ts, "int8_t"))
		return PV_TYPE_INT8;
	else if (!strcmp(ts, "int16_t"))
		return PV_TYPE_INT16;
	else if (!strcmp(ts, "int32_t"))
		return PV_TYPE_INT32;
	else if (!strcmp(ts, "float"))
		return PV_TYPE_FLOAT;
	else if (!strcmp(ts, "double"))
		return PV_TYPE_DOUBLE;
	else if (!strcmp(ts, "string"))
		return PV_TYPE_STRING;

	return PV_TYPE_UNDEF;
}

static enum pv_type pv_xml_get_array_type(const enum pv_type scalar_type)
{
	switch (scalar_type) {
	case PV_TYPE_UINT8:
		return PV_TYPE_UINT8_ARRAY;
	case PV_TYPE_UINT16:
		return PV_TYPE_UINT16_ARRAY;
	case PV_TYPE_UINT32:
		return PV_TYPE_UINT32_ARRAY;
	case PV_TYPE_INT8:
		return PV_TYPE_INT8_ARRAY;
	case PV_TYPE_INT16:
		return PV_TYPE_INT16_ARRAY;
	case PV_TYPE_INT32:
		return PV_TYPE_INT32_ARRAY;
	case PV_TYPE_FLOAT:
		return PV_TYPE_FLOAT_ARRAY;
	case PV_TYPE_DOUBLE:
		return PV_TYPE_DOUBLE_ARRAY;
	case PV_TYPE_STRING:
		return PV_TYPE_STRING;
	default:
		;
	}

	return PV_TYPE_UNDEF;
}

static void pv_xml_convert_string_to_val(union pv_dflt *const val,
					 const enum pv_type type)
{
	char *tmp = val->string;
	switch (type) {
	case PV_TYPE_STRING:
		return;		/* string: return, nothing to do */
	case PV_TYPE_INT8:
		val->s8 = strtol(tmp, NULL, 0);
		break;
	case PV_TYPE_INT16:
		val->s16 = strtol(tmp, NULL, 0);
		break;
	case PV_TYPE_INT32:
		val->s32 = strtol(tmp, NULL, 0);
		break;
	case PV_TYPE_INT64:
		val->s64 = strtol(tmp, NULL, 0);
		break;
	case PV_TYPE_UINT8:
		val->u8 = strtoul(tmp, NULL, 0);
		break;
	case PV_TYPE_UINT16:
		val->u16 = strtoul(tmp, NULL, 0);
		break;
	case PV_TYPE_UINT32:
		val->u32 = strtoul(tmp, NULL, 0);
		break;
	case PV_TYPE_UINT64:
		val->u64 = strtoul(tmp, NULL, 0);
		break;
	case PV_TYPE_FLOAT:
		val->f = strtod(tmp, NULL);
		break;
	case PV_TYPE_DOUBLE:
		val->d = strtod(tmp, NULL);
		break;
	case PV_TYPE_INT8_ARRAY:
	case PV_TYPE_INT16_ARRAY:
	case PV_TYPE_INT32_ARRAY:
	case PV_TYPE_UINT8_ARRAY:
	case PV_TYPE_UINT16_ARRAY:
	case PV_TYPE_UINT32_ARRAY:
	case PV_TYPE_FLOAT_ARRAY:
	case PV_TYPE_DOUBLE_ARRAY:
		val->ptr = NULL;
		break;
	case PV_TYPE_UNDEF:
		break;
	}
	free(tmp);
	return;
}

static int pv_xml_finalize_pv(struct pv *const pv)
{
	int err = 0;

	/* check for mandatory elements */
	assert(pv->id);
	if (pv->type == PV_TYPE_UNDEF) {
		dprintf(LEVEL_ERROR,
			"error: <pv> does not specify a valid type\n");
		err = EINVAL;
	}
	if (pv->offset == ~0) {
		dprintf(LEVEL_ERROR,
			"error: <pv> does not specify an offset\n");
		err = EINVAL;
	}
	if (pv->type == PV_TYPE_STRING && pv->size == 0) {
		dprintf(LEVEL_ERROR,
			"error: <pv> does not specify size of string\n");
		err = EINVAL;
	}
	/* check arraylen aka repeat */
	if (pv->arraylen > 1) {
		pv->type = pv_xml_get_array_type(pv->type);
	} else if (pv->arraylen != 1) {
		dprintf(LEVEL_ERROR,
			"error: <pv> does not specify valid repeat\n");
		err = EINVAL;
	}
	if (err)
		return err;

	/* set default size */
	if (pv->size == 0) {
		switch (pv->type) {
		case PV_TYPE_INT8:
		case PV_TYPE_UINT8:
		case PV_TYPE_INT8_ARRAY:
		case PV_TYPE_UINT8_ARRAY:	/* fallthrough */
			pv->size = 1;	/* FIXME: should be MAX(sizeof(uint8_t),sizeof(int8_t)) ? */
			break;
		case PV_TYPE_INT16:
		case PV_TYPE_UINT16:
		case PV_TYPE_INT16_ARRAY:
		case PV_TYPE_UINT16_ARRAY:	/* fallthrough */
			pv->size = 2;	/* FIXME: should be MAX(sizeof(uint16_t),sizeof(int16_t)) ? */
			break;
		case PV_TYPE_INT32:
		case PV_TYPE_UINT32:
		case PV_TYPE_INT32_ARRAY:
		case PV_TYPE_UINT32_ARRAY:
		case PV_TYPE_FLOAT:
		case PV_TYPE_FLOAT_ARRAY:	/* fallthrough */
			pv->size = 4;	/* FIXME: should be MAX(sizeof(uint32_t),sizeof(int32_t),sizeof(float)) ? */
			break;
		case PV_TYPE_INT64:
		case PV_TYPE_UINT64:
		case PV_TYPE_DOUBLE:
		case PV_TYPE_DOUBLE_ARRAY:	/* fallthrough */
			pv->size = 8;	/* FIXME: should be MAX(sizeof(uint64_t),sizeof(int64_t),sizeof(double)) ? */
			break;
		case PV_TYPE_STRING:
		case PV_TYPE_UNDEF:	/* fallthrough */
			break;
		}
	}

	/* set dflt (aka val), min, max according to type */
	if (pv->dflt.string)
		pv_xml_convert_string_to_val(&pv->dflt, pv->type);

	if (pv->has_max && pv->max.string)
		pv_xml_convert_string_to_val(&pv->max, pv->type);

	if (pv->has_min && pv->min.string)
		pv_xml_convert_string_to_val(&pv->min, pv->type);

	return err;
}

static void pv_xml_parse_data_init_pv(struct pv *const pv)
{
	memset(pv, 0, sizeof(struct pv));
	pv->offset = ~0;	/* check against this value, whether offset is set */
	pv->arraylen = 1;	/* a more reasonable default value */
}

static void pv_xml_parse_data_free_pv_content(struct pv *pv)
{
	free(pv->id);
	free(pv->name);
	free(pv->dflt.string);
	free(pv->min.string);
	free(pv->max.string);
	/*
	   free(pv->device); device is not implemented
	 */
}

static enum pv_xml_elem pv_xml_get_elem_name(const char *tag)
{
	int i, cmp;
	for (i = 0; i < PV_XML_ELEM_UNDEF; ++i) {
		cmp = strcmp(tag, pv_xml_tag[i]);
		if (cmp > 0)	/* tag is not in the _sorted_ list */
			break;
		if (cmp == 0)	/* tag is in the list */
			return i;
	}
	return PV_XML_ELEM_UNDEF;
}

static void XMLCALL pv_xml_start_handler(void *data, const XML_Char *el,
					 const XML_Char **attr)
{
	/*
	   FIXME: conversion from XML_Char to char!
	   XML_Char (defined in expat.h) is of type char for UTF-8 encoding. But
	   it is of type wchar_t or unsigned short for UTF-16 encoding. This
	   encoding is independent of the original encoding in the document.
	 */

	int i;
	enum pv_xml_elem elem;
	struct pv_xml_parse_data *pd = data;

	if (pd->err)
		return;

	if (pd->ignore) {	/* inside <pv> ignore unknown tags and their content */
		++pd->ignore;
		return;
	}

	if (pd->cur_elem != PV_XML_ELEM_UNDEF &&
	    pd->cur_elem != PV_XML_ELEM_PV) {	/* currently parsing a child of <pv>? */
		dprintf(LEVEL_ERROR,
			"(%s) error: unexpected opening tag <%s> in xml input file at line %d\n",
			__func__, el, (int)XML_GetCurrentLineNumber(*pd->parser));
		goto exit_0;
	}

	elem = pv_xml_get_elem_name(el);

	if (elem == PV_XML_ELEM_DEVICE) { /* <device> is not supported! */
		dprintf(LEVEL_ERROR, "error: <device> is not supported when using expat for parsing (at line %d of xml input file).\n"
			,(int)XML_GetCurrentLineNumber(*pd->parser));
	}

	if (!pd->pv_ptr->id && elem != PV_XML_ELEM_PV)	/* not inside <pv>, nothing to do */
		return;

	if (elem == PV_XML_ELEM_PV) {
		dprintf(LEVEL_DEBUG, "<pv>\n");
		for (i = 0; attr[i]; i += 2) {
			dprintf(LEVEL_DEBUG, "  %s = '%s'\n",
				attr[i], attr[i + 1]);
		}

		if (pd->pv_ptr->id) {	/* already inside a <pv> element? */
			dprintf(LEVEL_ERROR,
				"(%s) error: <pv> inside of <pv> at xml input file line %d\n",
				__func__,
				(int)XML_GetCurrentLineNumber(*pd->parser));
			goto exit_0;
		}
		for (i = 0; attr[i]; i += 2) {
			if (!strcmp(attr[i], "id"))
				pd->pv_ptr->id = strdup(attr[i + 1]);
			if (!strcmp(attr[i], "val"))
				pd->pv_ptr->dflt.string = strdup(attr[i + 1]);
		}
		/* is mandatory attribute 'id' missing or empty string? */
		if (!pd->pv_ptr->id || pd->pv_ptr->id[0] == 0x0) {
			dprintf(LEVEL_ERROR,
				"(%s) attribute 'id' is missing or empty at line %d of xml input file\n",
				__func__,
				(int)XML_GetCurrentLineNumber(*pd->parser));
			goto exit_0;
		}
	} else if (elem == PV_XML_ELEM_UNDEF) {
		pd->ignore = 1;
		dprintf(LEVEL_DEBUG, "  <%s> is ignored\n", el);
	} else {
		dprintf(LEVEL_DEBUG, "  <%s>\n", el);
	}

	free(pd->buf);
	pd->buf = NULL;

	pd->cur_elem = elem;
	return;
exit_0:
	XML_StopParser(*pd->parser, 0);
	pd->err = EINVAL;
	return;
}

static void XMLCALL pv_xml_end_handler(void *data, const XML_Char *el)
{
	/* FIXME: conversion from XML_Char to char! */
	int ret;
	struct pv_xml_parse_data *pd = data;

	if (pd->err)
		return;

	if (pd->ignore) {
		--pd->ignore;
		if (!pd->ignore)
			dprintf(LEVEL_DEBUG, "  </%s>\n", el);
		return;
	}

	if (!pd->pv_ptr->id)	/* not inside <pv> */
		return;

	assert(pd->cur_elem != PV_XML_ELEM_UNDEF);

	switch (pd->cur_elem) {
	case PV_XML_ELEM_PV:
		dprintf(LEVEL_DEBUG, "</pv>\n");
		/* do final checks and conversions */
		ret = pv_xml_finalize_pv(pd->pv_ptr);
		if (ret) {
			pd->err = EINVAL;
			XML_StopParser(*pd->parser, 0);
			break;
		}
		list_add_tail(pd->pv_list, pd->pv_ptr,
					    sizeof(struct pv));
		if (!pd->pv_list) {
			dprintf(LEVEL_CRIT,
				"(%s) error: could not allocate memory for list element.\n",
				__func__);
			pd->err = ENOMEM;
			XML_StopParser(*pd->parser, 0);
			break;
		}
		pv_xml_parse_data_init_pv(pd->pv_ptr);
		break;
	case PV_XML_ELEM_TYPE:
		dprintf(LEVEL_DEBUG, "  </type>\n");
		pd->pv_ptr->type = pv_xml_get_type(pd->buf);
		break;
	case PV_XML_ELEM_SIZE:
		dprintf(LEVEL_DEBUG, "  </size>\n");
		if (pd->buf)
			pd->pv_ptr->size = (size_t)strtoul(pd->buf, NULL, 0);
		break;
	case PV_XML_ELEM_REPEAT:
		dprintf(LEVEL_DEBUG, "  </repeat>\n");
		if (pd->buf)
			pd->pv_ptr->arraylen = (size_t)strtoul(pd->buf, NULL, 0);
		break;
	case PV_XML_ELEM_OFFSET:
		dprintf(LEVEL_DEBUG, "  </offset>\n");
		if (pd->buf)
			pd->pv_ptr->offset = (off_t)strtoul(pd->buf, NULL, 0);
		break;
	case PV_XML_ELEM_NAME:
		dprintf(LEVEL_DEBUG, "  </name>\n");
		if (pd->buf) {
			pd->pv_ptr->name = pd->buf;
			pd->buf = NULL;
		}
		break;
	case PV_XML_ELEM_MIN:
		dprintf(LEVEL_DEBUG, "  </min>\n");
		if (pd->buf) {
			pd->pv_ptr->has_min = 1;
			pd->pv_ptr->min.string = pd->buf;
			pd->buf = NULL;
		}
		break;
	case PV_XML_ELEM_MAX:
		dprintf(LEVEL_DEBUG, "  </max>\n");
		if (pd->buf) {
			pd->pv_ptr->max.string = pd->buf;
			pd->pv_ptr->has_max = 1;
			pd->buf = NULL;
		}
		break;
	case PV_XML_ELEM_UNDEF:
	default:
		;
	}

	/*
	 *  Change cur_elem according to XML structure when leaving <pv> itself
	 *  or one of its child elements
	 */
	if (pd->cur_elem == PV_XML_ELEM_PV)
		pd->cur_elem = PV_XML_ELEM_UNDEF;
	else
		pd->cur_elem = PV_XML_ELEM_PV;

	free(pd->buf);
	pd->buf = NULL;
}

static void XMLCALL pv_xml_content_handler(void *data, const XML_Char *s, int len)
{
	/* FIXME: conversion from XML_Char to char! */

	assert(len);
	size_t newlen, oldlen;
	struct pv_xml_parse_data *pd = data;

	if (pd->err)
		return;

	if (pd->cur_elem == PV_XML_ELEM_UNDEF ||
	    pd->cur_elem == PV_XML_ELEM_PV)
		return;

	oldlen = (!pd->buf) ? 0 : strlen(pd->buf);
	newlen = oldlen + len + 1;	/* 1 added, because space for additional terminating null byte is required */

	pd->buf = realloc(pd->buf, newlen * sizeof(char));

	if (!pd->buf) {
		dprintf(LEVEL_CRIT,
			"(%s) error: could not allocate memory for content string.\n",
			__func__);
		pd->err = ENOMEM;
		return;
	}

	if (!oldlen)	/* pd->buf points to newly allocated buffer, which must be initialised for strncat below */
		pd->buf[0] = 0x0;

	strncat(pd->buf, s, len);

	dprintf(LEVEL_DEBUG, "  current content: '%s'\n", pd->buf);
}

/**
 * pv_init_read_pvs - read in all pv definitions from an XML process description
 * @engine:
 * @xml_filename:
 *
 * Description:
 *
 * Return value;
 **/
int pv_init_read_pvs(struct pv_engine *engine, const char *xml_filename)
{
	int err = 0;
	int ret;
	int input;
	void *buff;
	struct pv curpv;
	struct pv_xml_parse_data parse_data = {
		.cur_elem = PV_XML_ELEM_UNDEF,
		.pv_ptr = &curpv,
	};

	pv_xml_parse_data_init_pv(parse_data.pv_ptr);

	input = open(xml_filename, O_RDONLY);
	if (input == -1) {
		err = -ENOENT;
		dprintf(LEVEL_CRIT, "(%s) could not open XML file \"%s\"\n",
			__func__, xml_filename);
		goto exit_0;
	}

	XML_Parser p = XML_ParserCreate(NULL);
	if (!p) {
		dprintf(LEVEL_CRIT,
			"(%s) error: could not allocate memory for parser.\n",
			__func__);
		err = -ENOMEM;
		goto exit_1;
	}
	parse_data.parser = &p;

	list_init(parse_data.pv_list);
	if (!parse_data.pv_list) {
		dprintf(LEVEL_CRIT,
			"(%s) error: could not allocate memory for list.\n",
			__func__);
		err = -ENOMEM;
		goto exit_1;
	}

	XML_SetElementHandler(p, pv_xml_start_handler, pv_xml_end_handler);
	XML_SetCharacterDataHandler(p, pv_xml_content_handler);
	XML_SetUserData(p, &parse_data);

	while (1) {
		int done;
		int len;

		buff = XML_GetBuffer(p, XMLBUFFERSIZE);
		if (!buff) {
			dprintf(LEVEL_CRIT,
				"(%s) error: could not allocate memory for parser (with XML_GetBuffer).\n",
				__func__);
			err = -ENOMEM;
			goto exit_2;
		}

		len = (int)read(input, buff, XMLBUFFERSIZE);
		if (len < 0) {
			err = errno;
			goto exit_2;
		}
		done = (len == 0);

		ret = XML_ParseBuffer(p, len, done);
		if (ret == XML_STATUS_ERROR) {	/* parser also returns XML_STATUS_ERROR if stopped by handler */
			dprintf(LEVEL_CRIT,
				"(%s) expat reports an error at line %d of '%s': %s\n",
				__func__,
				(int)XML_GetCurrentLineNumber(p),
				xml_filename,
				XML_ErrorString(XML_GetErrorCode(p)));
			err = -EINVAL;
			goto exit_2;
		}

		if (done)
			break;
	}

	/* Check of parse_data.err is unnecessary, because parser is aborted in this case */
	assert(parse_data.err == 0);

	if (!is_empty(parse_data.pv_list)) {	/* have there been pvs in file? */
		engine->pv_list = parse_data.pv_list;
	} else {
		dprintf(LEVEL_ERROR, "(%s) no pv entries in XML file found\n",
			__func__);
		err = -ENODEV;
	}
exit_2:
	free(parse_data.buf);
	if (err) {		/* in case of error free pvs including strings */
		pv_xml_parse_data_free_pv_content(parse_data.pv_ptr);
		if (!is_empty(parse_data.pv_list)) {
			ret = pv_destroy_free_pvs(parse_data.pv_list);
			if (ret) {
				dprintf(LEVEL_ERROR,
					"(%s) failed to destroy pv_list, err %d\n",
					__func__, err);
			}
		} else {
			list_del(parse_data.pv_list);
		}
	}
	XML_ParserFree(p);
exit_1:
	if (close(input) == -1)
		err = -errno;
exit_0:
	return err;
}

/* end of file pv_init_expat.c */
