/*
 * file: pv_init_libxml2.c
 * This file is part of the libpv library.
 *
 * Copyright 2005, 2009 Pengutronix
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 */

#include <libpv_config.h>

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>

#include <xmlconfig.h>

#include <libpv.h>

#include <pv_init.h>
#include <pv_destroy.h>

#define BUF_LEN		(127)
#define XPATH_LEN	(1023)

static int pv_init_init_pv_val(struct pv	*pv,
			       union pv_dflt	*val,
			       xmlconfig_t	*xmlcfg,
			       const char	*xpath)
{
	int		err = 0;
	unsigned long	tmp;
	double		dbl;

	if (pv->type == PV_TYPE_STRING) {
		val->string = calloc(1, pv->size);
		if (!val->string) {
			dprintf(LEVEL_CRIT, "(%s) error: could not get memory for id \"%s\"\n",
				__FUNCTION__, pv->id);
			return -ENOMEM;
		}
		err = xmlconfig_get_string(xmlcfg, xpath, val->string, pv->size, 1);
	} else if (pv->type == PV_TYPE_FLOAT ||
		   pv->type == PV_TYPE_DOUBLE) {
		err = xmlconfig_get_double(xmlcfg, xpath, &dbl, 1);
		if (err > 0) {
			switch (pv->type) {
			case PV_TYPE_FLOAT:
				val->f = dbl;
				break;
			case PV_TYPE_DOUBLE:
				val->d = dbl;
				break;
			default:
				break;
			}
		} else {
			val->u64 = 0;
		}
	} else if (pv->type != PV_TYPE_UNDEF) {
		err = xmlconfig_get_ulong(xmlcfg, xpath, &tmp, 1);	/* HACK: (mkl) */
		if (err > 0) {
			switch (pv->type) {
			case PV_TYPE_INT8:
				val->s8 = tmp;
				break;
			case PV_TYPE_UINT8:
				val->u8 = tmp;
				break;
			case PV_TYPE_INT16:
				val->s16 = tmp;
				break;
			case PV_TYPE_UINT16:
				val->u16 = tmp;
				break;
			case PV_TYPE_INT32:
				val->s32 = tmp;
				break;
			case PV_TYPE_UINT32:
				val->u32 = tmp;
				break;
			case PV_TYPE_INT64:
				val->s64 = tmp;
				break;
			case PV_TYPE_UINT64:
				val->u64 = tmp;
				break;

			case PV_TYPE_FLOAT_ARRAY:
			case PV_TYPE_DOUBLE_ARRAY:
			case PV_TYPE_INT8_ARRAY:
			case PV_TYPE_UINT8_ARRAY:
			case PV_TYPE_INT16_ARRAY:
			case PV_TYPE_UINT16_ARRAY:
			case PV_TYPE_INT32_ARRAY:
			case PV_TYPE_UINT32_ARRAY:
				/* FIXME */
				val->ptr = NULL;
				break;
			case PV_TYPE_FLOAT:
			case PV_TYPE_DOUBLE:
			case PV_TYPE_STRING:
			case PV_TYPE_UNDEF:
				break;
			}
		} else {
			val->u64 = 0;
		}
	}
	return err;
}

/**
 * pv_init_init_pv - read in pv information from an XML definition
 * @pv: pv to read data for
 * @xmlcfg: handle to XML description
 * @nr: pv index in XML description
 *
 * Description: Read in all relevant informations from the XML
 * process description for one process variable.
 *
 * Return value:
 **/
static int pv_init_init_pv(struct pv	*pv,
			   xmlconfig_t	*xmlcfg,
			   unsigned int	nr)
{
	char		buf[BUF_LEN + 1];
	char		xpath[XPATH_LEN + 1];
	unsigned long	tmp;
	int		err;
	char		*device_prefix;
	char		*device_index;
	char		*device_suffix;

	/* id */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/@id", nr);
	err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), nr);
	if (err <= 0) {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) pv %d not found, err %d\n",
			__FUNCTION__, nr, err);
		goto exit_0;
	}

	buf[BUF_LEN] = 0;
	pv->id = strdup(buf);	/* FIXME: limit size? strndup? (mkl) */
	if (!pv->id) {
		err = -ENOMEM;
		dprintf(LEVEL_CRIT, "(%s) error: could not get memory for id-buffer\n",
			__FUNCTION__);
		goto exit_0;
	}

	/* get repeat before type, because is repeat > 1 it is an array type */

	/* repeat */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/repeat", nr);
	err = xmlconfig_get_ulong(xmlcfg, xpath, &tmp, 1);	/* HACK: (mkl) */
	if (err == 0) {		/* no match, thus no repeat = 1 */
		pv->arraylen = 1;
	} else if (err < 0) {
		dprintf(LEVEL_CRIT, "(%s) repeat not found for id \"%s\", err %d\n",
			__FUNCTION__, pv->id, err);
		goto exit_2;
	} else {
		if (tmp < 1) {
			dprintf(LEVEL_CRIT, "(%s) repeat < 1 for id \"%s\"\n",
				__FUNCTION__, pv->id );
			goto exit_2;
		}

		pv->arraylen = tmp;
	}

	/* type */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/type", nr);
	err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), nr);
	if (err <= 0) {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) type not found for id \"%s\", err %d\n",
			__FUNCTION__, pv->id, err);
		goto exit_1;
	}

	buf[BUF_LEN] = 0;
/*
 * FIXME: to find the matching type id it would also possible
 * to use the pv_type_str table in pv_register.c
 */
	if (!strcmp(buf, "uint8_t")) {
		if (pv->arraylen == 1)
			pv->type = PV_TYPE_UINT8;
		else
			pv->type = PV_TYPE_UINT8_ARRAY;
	} else if (!strcmp(buf, "uint16_t")) {
		if (pv->arraylen == 1)
			pv->type = PV_TYPE_UINT16;
		else
			pv->type = PV_TYPE_UINT16_ARRAY;
	} else if (!strcmp(buf, "uint32_t")) {
		if (pv->arraylen == 1)
			pv->type = PV_TYPE_UINT32;
		else
			pv->type = PV_TYPE_UINT32_ARRAY;
	} else if (!strcmp(buf, "int8_t")) {
		if (pv->arraylen == 1)
			pv->type = PV_TYPE_INT8;
		else
			pv->type = PV_TYPE_INT8_ARRAY;
	} else if (!strcmp(buf, "int16_t")) {
		if (pv->arraylen == 1)
			pv->type = PV_TYPE_INT16;
		else
			pv->type = PV_TYPE_INT16_ARRAY;
	} else if (!strcmp(buf, "int32_t")) {
		if (pv->arraylen == 1)
			pv->type = PV_TYPE_INT32;
		else
			pv->type = PV_TYPE_INT32_ARRAY;
	} else if (!strcmp(buf, "float")) {
		if (pv->arraylen == 1)
			pv->type = PV_TYPE_FLOAT;
		else
			pv->type = PV_TYPE_FLOAT_ARRAY;
	} else if (!strcmp(buf, "double")) {
		if (pv->arraylen == 1)
			pv->type = PV_TYPE_DOUBLE;
		else
			pv->type = PV_TYPE_DOUBLE_ARRAY;
	} else if (!strcmp(buf,"string")) {
		pv->type = PV_TYPE_STRING;
	} else {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) unknown type, \"%s\" for id \"%s\"\n",
			__FUNCTION__, buf, pv->id);
		goto exit_1;
	}


	/* offset */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/offset", nr);
	err = xmlconfig_get_ulong(xmlcfg, xpath, &tmp, 1);	/* HACK: (mkl) */
	if (err <= 0) {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) offset not found for id \"%s\", err %d\n",
			__FUNCTION__, pv->id, err);
		goto exit_1;
	}
	pv->offset = tmp;


	/* name */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/name", nr);
	err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), 1);
	if (err <= 0) {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) name not found for id \"%s\", err %d\n",
			__FUNCTION__, pv->id, err);
		goto exit_1;
	}

	buf[BUF_LEN] = 0;
	pv->name = strdup(buf);	/* FIXME: limit size? strndup? (mkl) */
	if (!pv->name) {
		err = -ENOMEM;
		dprintf(LEVEL_CRIT, "(%s) error: could not get memory for name-buffer\n",
			__FUNCTION__);
		goto exit_1;
	}


	/* size */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/size", nr);
	err = xmlconfig_get_ulong(xmlcfg, xpath, &tmp, 1);	/* HACK: (mkl) */
	if (err == 0) {		/* no match, thus no size :) */
		switch (pv->type) {
		case PV_TYPE_INT8:
		case PV_TYPE_UINT8:
		case PV_TYPE_INT8_ARRAY:
		case PV_TYPE_UINT8_ARRAY:
			pv->size = 1;	/* FIXME: should be MAX(sizeof(uint8_t),sizeof(int8_t)) ? */
			break;
		case PV_TYPE_INT16:
		case PV_TYPE_UINT16:
		case PV_TYPE_INT16_ARRAY:
		case PV_TYPE_UINT16_ARRAY:
			pv->size = 2;	/* FIXME: should be MAX(sizeof(uint16_t),sizeof(int16_t)) ? */
			break;
		case PV_TYPE_INT32:
		case PV_TYPE_UINT32:
		case PV_TYPE_INT32_ARRAY:
		case PV_TYPE_UINT32_ARRAY:
		case PV_TYPE_FLOAT:
		case PV_TYPE_FLOAT_ARRAY:
			pv->size = 4;	/* FIXME: should be MAX(sizeof(uint32_t),sizeof(int32_t),sizeof(float)) ? */
			break;
		case PV_TYPE_INT64:
		case PV_TYPE_UINT64:
		case PV_TYPE_DOUBLE:
		case PV_TYPE_DOUBLE_ARRAY:
			pv->size = 8;	/* FIXME: should be MAX(sizeof(uint64_t),sizeof(int64_t),sizeof(double)) ? */
			break;
		case PV_TYPE_STRING:
			err = -EINVAL;
			dprintf(LEVEL_CRIT, "(%s) size not specified for string or undef type,"
				"id \"%s\"\n",
				__FUNCTION__, pv->id);
		case PV_TYPE_UNDEF:
			break;
		}
	} else if (err < 0) {
		dprintf(LEVEL_CRIT, "(%s) size not found for id \"%s\", err %d\n",
			__FUNCTION__, pv->id, err);
		goto exit_2;
	} else {
		pv->size = tmp;
	}


	/* default */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/@val", nr);
	err = pv_init_init_pv_val(pv, &pv->dflt, xmlcfg, xpath);
	if (err < 0) {
		dprintf(LEVEL_CRIT, "(%s) error while reading default value for pv %d, err %d\n",
			__FUNCTION__, nr, err);
		goto exit_2;
	}

	/* min */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/min", nr);
	err = pv_init_init_pv_val(pv, &pv->min, xmlcfg, xpath);
	if (err == -ERANGE) {
		/* tag exists with empty value */
		pv->has_min = 0;
	} else if (err < 0) {
		dprintf(LEVEL_CRIT, "(%s) error while reading min value for pv %d, err %d\n",
			__FUNCTION__, nr, err);
		goto exit_2;
	} else {
		pv->has_min = (err > 0);
	}

	/* max */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/max", nr);
	err = pv_init_init_pv_val(pv, &pv->max, xmlcfg, xpath);
	if (err == -ERANGE) {
		/* tag exists with empty value */
		pv->has_max = 0;
	} else if (err < 0) {
		dprintf(LEVEL_CRIT, "(%s) error while reading max value for pv %d, err %d\n",
			__FUNCTION__, nr, err);
		goto exit_2;
	} else {
		pv->has_max = (err > 0);
	}

	/* check if this pv is part of a <port> and get it's device path */
	snprintf(xpath, sizeof(xpath), "(//pv)[%i]/../../device", nr);
	err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), 0);
	if (err < 0) {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) error while searching device for pv %d, err %d\n",
			__FUNCTION__, nr, err);
		goto exit_2;
	}

	if (err == 1) {

		snprintf(xpath, sizeof(xpath), "(//pv)[%i][name(..)='portPvChannel']/../../device/@prefix", nr);
		buf[0]='\0';
		err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), nr);
		if (err < 0) {
			err = -EINVAL;
			dprintf(LEVEL_CRIT, "(%s) error while searching device prefix for pv %d, err %d\n",
				__FUNCTION__, nr, err);
			goto exit_2;
		}
		device_prefix = strdup(buf);	/* FIXME: limit size? strndup? (mkl) */

		snprintf(xpath, sizeof(xpath), "(//pv)[%i][name(..)='portPvChannel']/../../device/@index", nr);
		buf[0]='\0';
		err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), nr);
		if (err < 0) {
			err = -EINVAL;
			dprintf(LEVEL_CRIT, "(%s) error while searching device index for pv %d, err %d\n",
				__FUNCTION__, nr, err);
			goto exit_3;
		}
		device_index = strdup(buf);	/* FIXME: limit size? strndup? (mkl) */

		snprintf(xpath, sizeof(xpath), "(//pv)[%i][name(..)='portPvChannel']/../../device/@suffix", nr);
		buf[0]='\0';
		err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), nr);
		if (err < 0) {
			err = -EINVAL;
			dprintf(LEVEL_CRIT, "(%s) error while searching device suffix for pv %d, err %d\n",
				__FUNCTION__, nr, err);
			goto exit_4;
		}
		device_suffix = strdup(buf);	/* FIXME: limit size? strndup? (mkl) */

		snprintf(buf, sizeof(buf), "%s%s%s", device_prefix, device_index, device_suffix);
		pv->device = strdup(buf);

		free(device_suffix);
		free(device_index);
		free(device_prefix);

/*
 * Some devices (or sysfs entries) do only support one direction
 * So consider the given direction attribute
 */
		snprintf(xpath, sizeof(xpath), "(//pv)[%i][name(..)='portPvChannel']/../../device/@direction", nr);
		buf[0]='\0';
		err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), nr);
		if (err < 0) {
			err = -EINVAL;
			dprintf(LEVEL_CRIT, "(%s) error while searching device suffix for pv %d, err %d\n",
				__FUNCTION__, nr, err);
			goto exit_4;
		}
		if (!strcmp(buf,"in"))
			err = O_RDONLY;	/* this is only temporary var recycling ;-) */
		else {
			if (!strcmp(buf,"out"))
				err = O_WRONLY;
			else {
				if (!strcmp(buf,"inout"))
					err = O_RDWR;
			}
		}
		/*
		 * we have a device, open a file descriptor for it
		 */
		pv->device_file = open(pv->device, err); /* FIXME: make this nonblocking. */
		if (pv->device_file < 0) {
			fprintf(stderr, "warning: open(%s) for pv \"%s\" returned %i\n",
				pv->device, pv->id, pv->device_file
			);
		}
	} else {
		pv->device_file = -1;
	}


	/* FIXME: implement stuff for scale, unit, min, max (mkl)
	 * If someone does so, don't forget to update the documentation.
	 * Currently it states these tags as unsupported!
	 */

	return 0;

 exit_4:
	free(device_index);
 exit_3:
	free(device_prefix);
 exit_2:
	free(pv->name);
 exit_1:
	free(pv->id);
 exit_0:
	return err;
}

/**
 * pv_init_read_pvs - read in all pv definitions from an XML process description
 * @engine:
 * @xml_filename:
 *
 * Description:
 *
 * Return value;
 **/
int pv_init_read_pvs(struct pv_engine	*engine,
		     const char		*xml_filename)
{
	struct pv	pv;
	xmlconfig_t	*xmlcfg;
	int		no_pvs;
	int		err;
	int		ret;
	int		i;
	char		buf;

	/* analyse xml description file */
	xmlcfg = xmlconfig_init(xml_filename);
	if (!xmlcfg) {
		err = -ENOENT;
		dprintf(LEVEL_CRIT, "(%s) could not open XML file \"%s\"\n",
			__FUNCTION__, xml_filename);
		goto exit_0;
	}


	no_pvs = xmlconfig_get_string(xmlcfg, "//pv", &buf, 0, 0);
	if (no_pvs == 0) {
		err = -ENODEV;
		dprintf(LEVEL_WARN, "(%s) no pv entries in XML file found\n",
			__FUNCTION__);
		goto exit_1;
	} else if (no_pvs < 0) {
		err = no_pvs;
		dprintf(LEVEL_CRIT, "(%s) xmlconfig_get_string failed, err %d\n",
			__FUNCTION__, err);
		goto exit_1;
	}

	dprintf(LEVEL_INFO, "(%s) %d pv entries found\n",
		__FUNCTION__, no_pvs);

	/*
	 * create all process variables in a linked list
	 */
	list_init(engine->pv_list);
	for (i = 1; i <= no_pvs; i++) {
		memset(&pv, 0, sizeof(struct pv));
		err = pv_init_init_pv(&pv, xmlcfg, i);
		if (err)
			goto exit_2;

		list_add_tail(engine->pv_list, &pv, sizeof(struct pv));
	}

	err = xmlconfig_finish(xmlcfg);

	return err;

 exit_2:
	ret = pv_destroy_free_pvs(engine->pv_list);
    	if (ret)
		dprintf(LEVEL_ERROR, "(%s) failed to destroy pv_list, err %d\n",
			__FUNCTION__, ret);
 exit_1:
	ret = xmlconfig_finish(xmlcfg);
	if (ret)
		dprintf(LEVEL_ERROR, "(%s) xmlconfig_finish failed, err %d\n",
			__FUNCTION__, ret);
 exit_0:
	return err;
}

/* end of file pv_init_libxml2.c */
