/*
 * file: eventd_test.c
 * This file is part of the libpv library.
 *
 * Copyright 2005 Pengutronix
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 *
 * PURPOSE
 * This program should check as much features the eventd server provides
 *
 * It does all the tests 10 times to also check registering/deregistering whole clients.
 * It wouldn't be a good idea to start more than once eventd_test per eventd server,
 * because checks will be disturbed and fail!
 *
 * But to start more than one instance on one eventd server may discover other problems.
 * For example bad list handling (happens while developing...)!
 */

#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <pv_types.h>

struct pv_vars {
	unsigned long offset;
	int type;
	void *pointer;
	int registered;			/* 1=var is registered for change notifications */
	int notified;
	int changed;
};

static struct pv_vars *pv_vars=NULL;
static int  var_count=0;

/*
 * Simple thread to collect all notifications
 */
static void *notification_thread(void *p)
{
	pv_engine_t *pv_engine=(pv_engine_t*)p;
	void *var;
	int i;

	while(1) {
		if (pv_event_await_syncronous(pv_engine,&var) != 0)
			break;
		for (i=0;i<var_count;i++) {
			if (pv_vars[i].pointer == var) {
				pv_vars[i].notified++;
				break;
			}
		}
	}
	return(NULL);
}

/*
 * Tes: Try to register a notification to each process variable this
 * process image provides.
 *
 * FIXME: We should provide a library function to walk through all
 * vars present in this image. (moved to TODO.ptx)
 *
 * Currently we have to go the ugly way through internals
 */
static int check_pv_presense(struct pv_engine *engine)
{
	int		no_pvs=0;
	int		i,finished=0,fail;
	struct pv_list	*pv_list;
	struct pv *pv;
	pthread_t	thread=0;

	/*
	 * count list entries
	 */
	for (pv_list=engine->pv_list,goto_head(&pv_list); !finished; goto_next(&pv_list)) {
		no_pvs++;
		finished = is_last(pv_list);
	}

	if (no_pvs == 0) {
		fprintf(stderr, "(%s) no pv entries in this image found\n",__FUNCTION__);
		return(-1);
	}

	if ((pv_vars=calloc(no_pvs,sizeof(struct pv_vars))) == NULL) {
		fprintf(stderr, "(%s) no memory to store pv entries\n",__FUNCTION__);
		return(-1);
	}
	var_count=no_pvs;
	/*
	 * extract list entries
	 */
	for (i=finished=0,pv_list=engine->pv_list,goto_head(&pv_list); !finished; goto_next(&pv_list),i++) {
		finished = is_last(pv_list);
		pv = (struct pv*)list_get_element_data(pv_list);
		pv_vars[i].offset=(unsigned long)pv->offset;
		pv_vars[i].type=(int)pv->type;
		/*
		 * to be independend of any type, we calculate the
		 * pointer by ourself. Ugly internal...
		 */
		pv_vars[i].pointer=engine->store.map + pv->offset;
		pv_vars[i].registered=0; /* not notify registerd yet */
		pv_vars[i].notified=0;	/* not notified yet */
		pv_vars[i].changed=0;	/* no change yet */
	}
	/*
	 * register *all* vars for a change notification
	 */
	for (i=0;i<var_count;i++) {
		if (pv_event_register((pv_engine_t*)engine,pv_vars[i].pointer) < 0) {
			fprintf(stderr, "(%s) Cannot register var %p for change notification\n",
				__FUNCTION__,pv_vars[i].pointer);
			free(pv_vars);
			return(-1);
		}
		else
			pv_vars[i].registered++;
	}

/* ********************************************
   First test: Generate a change notification for each var.
   Check if all vars are notified after this.
   ******************************************** */

	fprintf(stdout,"TEST: Change all vars and wait for notification: ");

	if (pthread_create(&thread,NULL,notification_thread,engine) != 0) {
		fprintf(stderr,"(%s) Cannot create thread for notifying\n",__FUNCTION__);
		free(pv_vars);
		return(-1);
	}

	for (i=0;i<var_count;i++) {
		if (pv_event_generate((pv_engine_t*)engine,pv_vars[i].pointer) < 0) {
			fprintf(stderr,"(%s) Cannot generate a change event for var %p\n",
				__FUNCTION__,pv_vars[i].pointer);
			pthread_cancel(thread);
			free(pv_vars);
			return(-1);
		}
		else
			pv_vars[i].changed++;
	}
	/*
	 * FIXME: How long does it take to deliver all messages to the server and
	 * to receive it's answers?
	 */
	sleep(1);
	for (fail=0,i=0;i<var_count;i++) {
		if (pv_vars[i].notified != pv_vars[i].changed) {
			fprintf(stderr,"(%s) Var %p changed %d times, but notified %d times\n",
				__FUNCTION__,pv_vars[i].pointer,pv_vars[i].changed,pv_vars[i].notified);
			fail=1;
		}
	}
	pthread_cancel(thread);

	if (fail) {
		fprintf(stdout,"FAILED\n");
		free(pv_vars);
		return(-1);
	}
	else
		fprintf(stdout,"PASS\n");

/* ********************************************
   Second test: Deregister first, third and seventh var from change notifying,
   change all vars again, check if all remaining vars are notified after this.
   ******************************************** */

	fprintf(stdout,"TEST: Change all vars with deregistered 1st, 3rd and 7th var and wait for notification: ");

	if (var_count > 0) {
		if (pv_event_deregister((pv_engine_t*)engine,pv_vars[0].pointer) < 0) {
			fprintf(stderr, "(%s) Cannot deregister var %p from change notification\n",
				__FUNCTION__,pv_vars[0].pointer);
			free(pv_vars);
			return(-1);
		}
		else
			pv_vars[0].registered--;
	}
	if (var_count >= 3) {
		if (pv_event_deregister((pv_engine_t*)engine,pv_vars[2].pointer) < 0) {
			fprintf(stderr, "(%s) Cannot deregister var %p from change notification\n",
				__FUNCTION__,pv_vars[2].pointer);
			free(pv_vars);
			return(-1);
		}
		else
			pv_vars[2].registered--;
	}
	if (var_count >= 7) {
		if (pv_event_deregister((pv_engine_t*)engine,pv_vars[6].pointer) < 0) {
			fprintf(stderr, "(%s) Cannot deregister var %p from change notification\n",
				__FUNCTION__,pv_vars[6].pointer);
			free(pv_vars);
			return(-1);
		}
		else
			pv_vars[6].registered--;
	}

	if (pthread_create(&thread,NULL,notification_thread,engine) != 0) {
		fprintf(stderr,"(%s) Cannot create thread for notifying\n",__FUNCTION__);
		free(pv_vars);
		return(-1);
	}

	for (i=0;i<var_count;i++) {
		if (pv_event_generate((pv_engine_t*)engine,pv_vars[i].pointer) < 0) {
			fprintf(stderr,"(%s) Cannot generate a change event for var %p\n",
				__FUNCTION__,pv_vars[i].pointer);
			pthread_cancel(thread);
			free(pv_vars);
			return(-1);
		}
		else
			pv_vars[i].changed++;
	}
	/*
	 * FIXME: How long does it take to deliver all messages to the server and
	 * to receive it's answers?
	 */
	sleep(1);
	for (fail=0,i=0;i<var_count;i++) {
		if (pv_vars[i].notified != pv_vars[i].changed) {
			if ((i==0) || (i==2) || (i==6)) {
				if (pv_vars[i].notified != (pv_vars[i].changed-1)) {
					fprintf(stderr,"(%s) Var %p changed %d times, but notified %d times. Should be %d.\n",
						__FUNCTION__,pv_vars[i].pointer,pv_vars[i].changed,pv_vars[i].notified,
						pv_vars[i].changed-1);
					fail=1;
				}
			}
			else {
				fprintf(stderr,"(%s) Var %p changed %d times, but notified %d times\n",
					__FUNCTION__,pv_vars[i].pointer,pv_vars[i].changed,pv_vars[i].notified);
				fail=1;
			}
		}
	}
	pthread_cancel(thread);

	if (fail) {
		fprintf(stdout,"FAILED\n");
		free(pv_vars);
		return(-1);
	}
	else
		fprintf(stdout,"PASS\n");

/* ********************************************
   Third test: Deregister all, change all and see if notified (should not)
   ******************************************** */

	fprintf(stdout,"TEST: Deregister all, change and test if notified: ");

	for (i=0;i<var_count;i++) {
		if (pv_event_deregister((pv_engine_t*)engine,pv_vars[i].pointer) < 0) {
			fprintf(stderr, "(%s) Cannot deregister var %p from change notification\n",
				__FUNCTION__,pv_vars[i].pointer);
			free(pv_vars);
			return(-1);
		}
		else
			pv_vars[i].registered--;

		pv_vars[i].notified=pv_vars[i].changed=0;	/* reset counter */
	}

	if (pthread_create(&thread,NULL,notification_thread,engine) != 0) {
		fprintf(stderr,"(%s) Cannot create thread for notifying\n",__FUNCTION__);
		free(pv_vars);
		return(-1);
	}

	for (i=0;i<var_count;i++) {
		if (pv_event_generate((pv_engine_t*)engine,pv_vars[i].pointer) < 0) {
			fprintf(stderr,"(%s) Cannot generate a change event for var %p\n",
				__FUNCTION__,pv_vars[i].pointer);
			pthread_cancel(thread);
			free(pv_vars);
			return(-1);
		}
		else
			pv_vars[i].changed++;
	}
	/*
	 * FIXME: How long does it take to deliver all messages to the server and
	 * to receive it's answers?
	 */
	sleep(1);
	for (fail=0,i=0;i<var_count;i++) {
		if (pv_vars[i].notified != (pv_vars[i].changed-1)) {
			fprintf(stderr,"(%s) Var %p changed %d times, but notified %d times. Should be %d\n",
				__FUNCTION__,pv_vars[i].pointer,pv_vars[i].changed,
				pv_vars[i].notified,pv_vars[i].changed-1);
			fail=1;
		}
	}
	pthread_cancel(thread);

	if (fail) {
		fprintf(stdout,"FAILED\n");
		free(pv_vars);
		return(-1);
	}
	else
		fprintf(stdout,"PASS\n");

	free(pv_vars);
	return(0);
}

int main(int argc, char *argv[])
{
	pv_engine_t 		pv_engine;
	pv_engine_attr_t	attr;
	int		err=EXIT_SUCCESS;
	int n=10;

	while (n) {
		pv_vars=NULL;
		var_count=0;

		if (pv_engine_attr_init_from_cmdline(&attr, argc, argv))
			exit (EXIT_FAILURE);
		fprintf(stdout,"Register client %d to server.\n",getpid());
		if (pv_init(&pv_engine, &attr))
			exit (EXIT_FAILURE);

		if (check_pv_presense((struct pv_engine*)&pv_engine))
			err=EXIT_FAILURE;

		fprintf(stdout,"Deregister client %d from server.\n",getpid());
		pv_destroy(&pv_engine);
		if (err == EXIT_FAILURE)
			break;
		n--;
	}

	return(err);
}

/* end of file eventd_test.c */
